/* Examples and Documentation for
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 * 
 * 
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#pragma once

////// NOTE OP-E: you don't need to change anything in this file

#include <scgms/rtl/FilterLib.h>
#include <scgms/rtl/referencedImpl.h>
#include <scgms/rtl/UILib.h>

#include "descriptor.h"

#pragma warning( push )
#pragma warning( disable : 4250 ) // C4250 - 'class1' : inherits 'class2::member' via dominance

/*
 * Base class for all control models for your OP-E work
 */
class CBase_Control_Model : public scgms::CBase_Filter, public virtual scgms::IDiscrete_Model {

	protected:
		// stored segment ID
		uint64_t mSegment_id = scgms::Invalid_Segment_Id;

		// we have to hold the current timestamp
		double mCurrent_Time = 0;

		// ID of a model
		GUID mModel_Id;
		// ID of bolus insulin signal
		GUID mBolus_Signal;
		// ID of basal insulin rate signal
		GUID mBasal_Signal;

	protected:
		// scgms::CBase_Filter iface implementation
		virtual HRESULT Do_Execute(scgms::UDevice_Event event) override final {
			if (!On_Event_Execute(event))
				return S_OK;
			return mOutput.Send(event);
		}
		virtual HRESULT Do_Configure(scgms::SFilter_Configuration configuration, refcnt::Swstr_list& error_description) override final {
			return S_OK;
		}

		// emits a signal level at a given time
		bool Emit_Signal_Level(double level, double time, const GUID& signal_id) {
			scgms::UDevice_Event evt{ scgms::NDevice_Event_Code::Level };
			evt.signal_id() = signal_id;
			evt.device_id() = mModel_Id;
			evt.device_time() = time;
			evt.segment_id() = mSegment_id;
			evt.level() = level;
			return Succeeded(mOutput.Send(evt));
		}

		// emits a bolus insulin signal
		bool Emit_Bolus(double level, double time) {
			return Emit_Signal_Level(level, time, mBolus_Signal);
		}

		// emits a basal insulin rate change signal
		bool Emit_Basal_Change(double level, double time) {
			return Emit_Signal_Level(level, time, mBasal_Signal);
		}

		// called upon event receipt (Do_Execute)
		virtual bool On_Event_Execute(scgms::UDevice_Event& event) = 0;
		// called upon Step method call
		virtual void Advance_Model(double time_advance_delta) = 0;

	public:
		CBase_Control_Model(scgms::IFilter* output, const GUID& model_id, const GUID& bolus_signal, const GUID& basal_signal) : scgms::CBase_Filter(output), mModel_Id(model_id), mBolus_Signal(bolus_signal), mBasal_Signal(basal_signal) {
			//
		}
		virtual ~CBase_Control_Model() = default;

		// scgms::IDiscrete_Model iface
		virtual HRESULT IfaceCalling Initialize(const double new_current_time, const uint64_t segment_id) override final {
			mCurrent_Time = new_current_time;
			mSegment_id = segment_id;
			return S_OK;
		}

		virtual HRESULT IfaceCalling Step(const double time_advance_delta) override final {
			if (time_advance_delta > 0.0) {
				// really advance the model here
				Advance_Model(time_advance_delta);

				// advance the timestamp
				mCurrent_Time += time_advance_delta;
			}

			return S_OK;
		}
};

#pragma warning( pop )
