/* Examples and Documentation for
 * SmartCGMS - continuous glucose monitoring and controlling framework
 * https://diabetes.zcu.cz/
 *
 * Copyright (c) since 2018 University of West Bohemia.
 *
 * Contact:
 * diabetes@mail.kiv.zcu.cz
 * Medical Informatics, Department of Computer Science and Engineering
 * Faculty of Applied Sciences, University of West Bohemia
 * Univerzitni 8, 301 00 Pilsen
 * Czech Republic
 * 
 * 
 * Purpose of this software:
 * This software is intended to demonstrate work of the diabetes.zcu.cz research
 * group to other scientists, to complement our published papers. It is strictly
 * prohibited to use this software for diagnosis or treatment of any medical condition,
 * without obtaining all required approvals from respective regulatory bodies.
 *
 * Especially, a diabetic patient is warned that unauthorized use of this software
 * may result into severe injure, including death.
 *
 *
 * Licensing terms:
 * Unless required by applicable law or agreed to in writing, software
 * distributed under these license terms is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#pragma once

////// NOTE OP-E: you don't need to change anything in this file

#include <scgms/iface/DeviceIface.h>
#include <scgms/rtl/SolverLib.h>
#include <scgms/utils/descriptor_utils.h>
#include <scgms/rtl/manufactory.h>
#include <scgms/rtl/guid.h>

#include <map>
#include <vector>

// macro for parameters structure definition
#define DEFINE_PARAMETERS_STRUCTURE(_type_name, _parameter_count, _structure_block) const size_t parameters_count = _parameter_count; struct _type_name { union { struct _structure_block; double vector[_parameter_count]; }; }

// helper structure for model storing
struct TModel_Descriptor_Helper
{
	GUID id;
	scgms::NModel_Flags flags;
	std::wstring description;
	std::wstring db_table_name;

	size_t total_number_of_parameters;
	size_t number_of_segment_specific_parameters;
	std::vector<scgms::NModel_Parameter_Value> parameter_types;

	std::vector<std::wstring> _parameter_ui_names_proxy;
	std::vector<const wchar_t*> parameter_ui_names;

	std::vector<double> lower_bound;
	std::vector<double> default_values;
	std::vector<double> upper_bound;

	std::vector<GUID> calculated_signal_ids;
	std::vector<GUID> reference_signal_ids;

	std::vector<scgms::TSignal_Descriptor> signal_descriptors;
};

// helper structure for model parameters
struct TModel_Parameter_Helper
{
	const std::wstring name;
	const scgms::NModel_Parameter_Value type;
	const double lower_bound;
	const double default_value;
	const double upper_bound;
};

// template model factory
template<typename TModel_Class>
HRESULT Discrete_Model_Factory(scgms::IModel_Parameter_Vector* parameters, scgms::IFilter* output, scgms::IDiscrete_Model** model) {
	return Manufacture_Object<TModel_Class>(model, parameters, output);
}

// model factory function pointer
using TModel_Factory = HRESULT(*)(scgms::IModel_Parameter_Vector*, scgms::IFilter*, scgms::IDiscrete_Model**);

extern std::vector<TModel_Descriptor_Helper> Defined_Models_Storage;
extern std::map<GUID, TModel_Factory> Model_Factories;

// helper function to define a discrete model
template<typename TModel_Class>
void Describe_Discrete_Model(
	const GUID& model_id,
	const std::wstring& model_name,
	const std::vector<TModel_Parameter_Helper>& parameters,
	const std::vector<scgms::TSignal_Descriptor>& signals)
{
	// create model descriptor helper to fill later
	auto& hlp = Defined_Models_Storage.emplace_back(TModel_Descriptor_Helper{
		model_id,
		scgms::NModel_Flags::Discrete_Model,
		model_name,
		L"",
		parameters.size(),
		parameters.size()
		});

	// fill parameters
	for (auto& param : parameters) {
		hlp.parameter_types.push_back(param.type);
		hlp._parameter_ui_names_proxy.push_back(param.name);
		hlp.lower_bound.push_back(param.lower_bound);
		hlp.default_values.push_back(param.default_value);
		hlp.upper_bound.push_back(param.upper_bound);
	}

	// transcribe model parameter names
	for (auto& pname : hlp._parameter_ui_names_proxy) {
		hlp.parameter_ui_names.push_back(pname.c_str());
	}

	// create signal descriptors
	for (auto& signal : signals) {
		hlp.calculated_signal_ids.push_back(signal.id);
		hlp.reference_signal_ids.push_back(signal.id); // TODO: descriptor should contain this information, or we need a different means of getting it in a generic way

		hlp.signal_descriptors.push_back(signal);
	}

	// register model factory
	Model_Factories[model_id] = &Discrete_Model_Factory<TModel_Class>;
}

void init_entities();
